@@ -53,33 +53,57 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
+#ifdef _WIN32
+#define WIN32_LEAN_AND_MEAN
+#include <windows.h>
+#else // _WIN32
 #include <X11/Xlib.h>
 #include <X11/keysym.h>
+#endif // _WIN32
 #ifdef USE_GLU
 #include <GL/glu.h>
 #endif
 #include <GL/gl.h>
 
+#ifdef _WIN32
+// disable double->float conversion warnings
+#pragma warning(disable : 4244)
+#pragma warning(disable : 4305)
+#else // _WIN32
 typedef void (*__GLXextFuncPtr)(void);
 extern __GLXextFuncPtr glXGetProcAddressARB(const GLubyte *);
 #define GLX_ARB_get_proc_address 1
 
 #include <GL/glx.h>
+#endif // _WIN32
 
 #define INT_PTR ptrdiff_t
 #include <GL/glATI.h>
 
+#ifdef _WIN32
+#include <GL/wglATI.h>
+#include <time.h>
+#else // _WIN32
 #include <sys/time.h>
 #include <unistd.h>
+#endif // _WIN32
 
 /* return current time (in seconds) */
 static int
 current_time(void)
 {
+#ifdef _WIN32
+    clock_t tv;
+
+    tv = clock();
+
+    return (tv / CLOCKS_PER_SEC);
+#else // _WIN32
    struct timeval tv;
    struct timezone tz;
    (void) gettimeofday(&tv, &tz);
    return (int) tv.tv_sec;
+#endif // _WIN32
 }
 
 #ifndef M_PI
@@ -98,6 +122,12 @@ static PFNGLRENDERBUFFERSTORAGEEXTPROC glRenderbufferStorageEXT;
 static PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC glFramebufferRenderbufferEXT;
 static PFNGLDELETEFRAMEBUFFERSEXTPROC glDeleteFramebuffersEXT;
 static PFNGLDELETERENDERBUFFERSEXTPROC glDeleteRenderbuffersEXT;
+#ifdef _WIN32
+static PFNWGLCREATEPBUFFERARBPROC wglCreatePbufferARB;
+static PFNWGLDESTROYPBUFFERARBPROC wglDestroyPbufferARB;
+static PFNWGLGETPBUFFERDCARBPROC wglGetPbufferDCARB;
+static PFNWGLRELEASEPBUFFERDCARBPROC wglReleasePbufferDCARB;
+#endif // _WIN32
 
 
 /* dimensions of texture and pbuffer */
@@ -105,12 +135,21 @@ static PFNGLDELETERENDERBUFFERSEXTPROC glDeleteRenderbuffersEXT;
 #define TEXTURE_BORDER_WIDTH  3
 #define PBUFFER_WIDTH_HEIGHT  TEXTURE_SIZE_XY - 2*TEXTURE_BORDER_WIDTH
 
+#ifdef _WIN32
+static HWND hWnd;
+static HDC hDC;
+static HDC hDC_pbuf;
+static HGLRC hGLRC;
+static HGLRC hGLRC_pbuf;
+static HPBUFFERARB pbuf;
+#else // _WIN32
 /* for simplicity, i take global var. for X and GLX resources */
 static Display *dpy        = NULL;
 static Window win          = 0;
 static GLXPbuffer pbuf     = 0;
 static GLXContext ctx_win  = NULL,
                   ctx_pbuf = NULL;
+#endif // _WIN32
 
 /* gears stuff */
 static GLfloat view_rotx = 20.0,
@@ -329,7 +368,11 @@ static void drawGears(void)
    }
    else
    {
+#ifdef _WIN32
+      wglMakeCurrent(hDC_pbuf, hGLRC_pbuf);
+#else
       glXMakeCurrent(dpy, pbuf, ctx_pbuf);
+#endif // _WIN32
    }
 
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
@@ -470,7 +513,11 @@ static void drawCube(void)
     }
     else
     {
+#ifdef _WIN32
+        wglMakeCurrent(hDC, hGLRC);
+#else
         glXMakeCurrent(dpy, win, ctx_win);
+#endif // _WIN32
     }
 
     glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
@@ -553,7 +600,11 @@ static void drawCube(void)
         rotZ -= 7200.0f;
     }
 
+#ifdef _WIN32
+    SwapBuffers(hDC);
+#else // _WIN32
     glXSwapBuffers(dpy, win);
+#endif // _WIN32
 }
 
 
@@ -576,7 +627,11 @@ static void init(void)
    if (!use_fbo)
    {
       /* stuff for window */
+#ifdef _WIN32
+      wglMakeCurrent(hDC, hGLRC);
+#else
       glXMakeCurrent(dpy, win, ctx_win);
+#endif
 
       setCommonWindowState();
    }
@@ -594,7 +649,11 @@ static void init(void)
    if (!use_fbo)
    {
       /* stuff for pbuffer */
+#ifdef _WIN32
+      wglMakeCurrent(hDC_pbuf, hGLRC_pbuf);
+#else
       glXMakeCurrent(dpy, pbuf, ctx_pbuf);
+#endif
 
       setCommonGearsState();
    }
@@ -621,12 +680,143 @@ static void init(void)
    glEnable(GL_NORMALIZE);
 }
 
+#ifdef _WIN32
+LRESULT CALLBACK WndProc(   HWND    hWnd,           // Handle For This Window
+                            UINT    uMsg,           // Message For This Window
+                            WPARAM  wParam,         // Additional Message Information
+                            LPARAM  lParam)         // Additional Message Information
+{
+    switch (uMsg)
+    {
+        
+    case WM_CREATE:
+        return 0;
+        
+    case WM_CLOSE:
+        PostQuitMessage( 0 );
+        return 0;
+        
+    case WM_DESTROY:
+        return 0;
+
+    case WM_SIZE:
+	    /* track window size changes */
+        if (!use_fbo)
+        {
+            wglMakeCurrent(hDC, hGLRC);
+        }
+
+        reshapeCube(LOWORD(lParam), HIWORD(lParam));
+        return 0;
+
+    case WM_KEYDOWN:
+        switch ( wParam )
+        {
+        case VK_LEFT:
+            view_roty += 5.0;
+            break;
+        case VK_RIGHT:
+            view_roty -= 5.0;
+            break;
+        case VK_UP:
+            view_rotx += 5.0;
+            break;
+        case VK_DOWN:
+            view_rotx -= 5.0;
+            break;
+        case VK_ESCAPE:
+            PostQuitMessage(0);
+            return 0;
+            
+        }
+        return 0;
+    }
+    // Pass All Unhandled Messages To DefWindowProc
+    return DefWindowProc(hWnd,uMsg,wParam,lParam);
+}
+#endif // _WIN32
 
 /*
  * Create an RGBA, double-buffered window.
  */
 static void make_window(const char *name, int x, int y, int width, int height)
 {
+#ifdef _WIN32
+    static const PIXELFORMATDESCRIPTOR pfd =
+    {
+        sizeof(PIXELFORMATDESCRIPTOR),  /* size */
+        1,				            /* version */
+        PFD_SUPPORT_OPENGL |
+        PFD_DRAW_TO_WINDOW |
+        PFD_DOUBLEBUFFER,		    /* support double-buffering */
+        PFD_TYPE_RGBA,			    /* color type */
+        16,				            /* prefered color depth */
+        0, 0, 0, 0, 0, 0,		    /* color bits (ignored) */
+        0,				            /* no alpha buffer */
+        0,				            /* alpha bits (ignored) */
+        0,				            /* no accumulation buffer */
+        0, 0, 0, 0,			        /* accum bits (ignored) */
+        16,				            /* depth buffer */
+        1,				            /* no stencil buffer */
+        0,				            /* no auxiliary buffers */
+        PFD_MAIN_PLANE,			    /* main layer */
+        0,				            /* reserved */
+        0, 0, 0,			        /* no layer, visible, damage masks */
+    };
+    int pixelFormat;
+    static char szAppName[] = "FGL glxgears";
+    HINSTANCE hInstance = GetModuleHandle(NULL);
+    WNDCLASS wc;
+
+    wc.style   = CS_HREDRAW | CS_VREDRAW;
+    wc.lpfnWndProc      = (WNDPROC)WndProc;
+    wc.cbClsExtra       = 0;
+    wc.cbWndExtra       = 0;
+    wc.hInstance        = hInstance;
+    wc.hIcon            = NULL;
+    wc.hCursor          = LoadCursor(NULL, IDC_ARROW);
+    wc.hbrBackground    = (HBRUSH)(COLOR_WINDOW+1);
+    wc.lpszMenuName     = NULL;
+    wc.lpszClassName    = szAppName;
+
+    RegisterClass(&wc);
+
+    /* create window */
+    hWnd = CreateWindow(
+        szAppName, name,
+        WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
+        x, y, width, height,
+        NULL, NULL, hInstance, NULL);
+
+    /* display window */
+    ShowWindow(hWnd, SW_SHOWNORMAL);
+    UpdateWindow(hWnd);
+
+    /* initialize OpenGL rendering */
+    hDC = GetDC(hWnd);
+
+    pixelFormat = ChoosePixelFormat(hDC, &pfd);
+    if (pixelFormat == 0)
+    {
+        printf("Error: couldn't get an RGBA, Double-buffered pixel format\n");
+        exit(1);
+    }
+
+    if (SetPixelFormat(hDC, pixelFormat, &pfd) != TRUE) 
+    {
+        printf("Error: SetPixelFormat failed\n");
+        exit(1);
+    }
+
+    hGLRC = wglCreateContext(hDC);
+    if (!hGLRC)
+    {
+        printf("Error: wglCreateContext failed\n");
+        exit(1);
+    }
+
+    wglMakeCurrent(hDC, hGLRC);
+#else // _WIN32
    int attrib[] = { GLX_RGBA,
 		    GLX_RED_SIZE, 1,
 		    GLX_GREEN_SIZE, 1,
@@ -689,21 +879,27 @@ static void make_window(const char *name, int x, int y, int width, int height)
    XFree(visinfo);
 
    glXMakeCurrent(dpy, win, ctx_win);
+#endif // _WIN32
 }
 
+#ifdef _WIN32
+#define GET_PROC_ADDRESS wglGetProcAddress
+#else // WIN32
+#define GET_PROC_ADDRESS glXGetProcAddressARB
+#endif // _WIN32
 
 static void
 make_fbo(int width, int height)
 {
-   glGenFramebuffersEXT = (PFNGLGENFRAMEBUFFERSEXTPROC)glXGetProcAddressARB("glGenFramebuffersEXT");
-   glBindFramebufferEXT = (PFNGLBINDFRAMEBUFFEREXTPROC)glXGetProcAddressARB("glBindFramebufferEXT");
-   glFramebufferTexture2DEXT = (PFNGLFRAMEBUFFERTEXTURE2DEXTPROC)glXGetProcAddressARB("glFramebufferTexture2DEXT");
-   glGenRenderbuffersEXT = (PFNGLGENRENDERBUFFERSEXTPROC)glXGetProcAddressARB("glGenRenderbuffersEXT");
-   glBindRenderbufferEXT = (PFNGLBINDRENDERBUFFEREXTPROC)glXGetProcAddressARB("glBindRenderbufferEXT");
-   glRenderbufferStorageEXT = (PFNGLRENDERBUFFERSTORAGEEXTPROC)glXGetProcAddressARB("glRenderbufferStorageEXT");
-   glFramebufferRenderbufferEXT = (PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC)glXGetProcAddressARB("glFramebufferRenderbufferEXT");
-   glDeleteFramebuffersEXT = (PFNGLDELETEFRAMEBUFFERSEXTPROC)glXGetProcAddressARB("glDeleteFramebuffersEXT");
-   glDeleteRenderbuffersEXT = (PFNGLDELETERENDERBUFFERSEXTPROC)glXGetProcAddressARB("glDeleteRenderbuffersEXT");
+   glGenFramebuffersEXT = (PFNGLGENFRAMEBUFFERSEXTPROC)GET_PROC_ADDRESS((const GLubyte*)"glGenFramebuffersEXT");
+   glBindFramebufferEXT = (PFNGLBINDFRAMEBUFFEREXTPROC)GET_PROC_ADDRESS((const GLubyte*)"glBindFramebufferEXT");
+   glFramebufferTexture2DEXT = (PFNGLFRAMEBUFFERTEXTURE2DEXTPROC)GET_PROC_ADDRESS((const GLubyte*)"glFramebufferTexture2DEXT");
+   glGenRenderbuffersEXT = (PFNGLGENRENDERBUFFERSEXTPROC)GET_PROC_ADDRESS((const GLubyte*)"glGenRenderbuffersEXT");
+   glBindRenderbufferEXT = (PFNGLBINDRENDERBUFFEREXTPROC)GET_PROC_ADDRESS((const GLubyte*)"glBindRenderbufferEXT");
+   glRenderbufferStorageEXT = (PFNGLRENDERBUFFERSTORAGEEXTPROC)GET_PROC_ADDRESS((const GLubyte*)"glRenderbufferStorageEXT");
+   glFramebufferRenderbufferEXT = (PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC)GET_PROC_ADDRESS((const GLubyte*)"glFramebufferRenderbufferEXT");
+   glDeleteFramebuffersEXT = (PFNGLDELETEFRAMEBUFFERSEXTPROC)GET_PROC_ADDRESS((const GLubyte*)"glDeleteFramebuffersEXT");
+   glDeleteRenderbuffersEXT = (PFNGLDELETERENDERBUFFERSEXTPROC)GET_PROC_ADDRESS((const GLubyte*)"glDeleteRenderbuffersEXT");
 
    /* make framebuffer */
    glGenFramebuffersEXT(1, &gearsFB);
@@ -723,6 +919,73 @@ make_fbo(int width, int height)
  */
 static void make_pbuffer(int width, int height)
 {
+#ifdef _WIN32
+    static const PIXELFORMATDESCRIPTOR pfd =
+    {
+        sizeof(PIXELFORMATDESCRIPTOR),  /* size */
+        1,				            /* version */
+        PFD_SUPPORT_OPENGL |
+        PFD_DRAW_TO_WINDOW,
+        PFD_TYPE_RGBA,			    /* color type */
+        16,				            /* prefered color depth */
+        0, 0, 0, 0, 0, 0,		    /* color bits (ignored) */
+        0,				            /* no alpha buffer */
+        0,				            /* alpha bits (ignored) */
+        0,				            /* no accumulation buffer */
+        0, 0, 0, 0,			        /* accum bits (ignored) */
+        16,				            /* depth buffer */
+        1,				            /* no stencil buffer */
+        0,				            /* no auxiliary buffers */
+        PFD_MAIN_PLANE,			    /* main layer */
+        0,				            /* reserved */
+        0, 0, 0,			        /* no layer, visible, damage masks */
+    };
+    int pixelFormat;
+
+    pixelFormat = ChoosePixelFormat(hDC, &pfd);
+    if (pixelFormat == 0)
+    {
+        printf("Error: couldn't get an RGBA, Double-buffered pixel format\n");
+        exit(1);
+    }
+
+    wglCreatePbufferARB = (PFNWGLCREATEPBUFFERARBPROC)GET_PROC_ADDRESS("wglCreatePbufferARB");
+    wglDestroyPbufferARB = (PFNWGLDESTROYPBUFFERARBPROC)GET_PROC_ADDRESS("wglDestroyPbufferARB");
+    wglGetPbufferDCARB = (PFNWGLGETPBUFFERDCARBPROC)GET_PROC_ADDRESS("wglGetPbufferDCARB");
+    wglReleasePbufferDCARB = (PFNWGLRELEASEPBUFFERDCARBPROC)GET_PROC_ADDRESS("wglReleasePbufferDCARB");
+
+    pbuf = wglCreatePbufferARB(hDC, pixelFormat, width, height, NULL);
+    if(!pbuf)
+    {
+        printf("Error: wglCreatePbufferARB failed\n");
+        exit(1);
+    }
+    hDC_pbuf = wglGetPbufferDCARB(pbuf);
+    if(!hDC_pbuf)
+    {
+        printf("Error: wglCreatePbufferARB failed\n");
+        exit(1);
+    }
+
+    if (SetPixelFormat(hDC_pbuf, pixelFormat, &pfd) != TRUE) 
+    {
+        printf("Error: SetPixelFormat failed\n");
+        exit(1);
+    }
+
+    hGLRC_pbuf = wglCreateContext(hDC_pbuf);
+    if (!hGLRC_pbuf)
+    {
+        printf("Error: wglCreateContext failed\n");
+        exit(1);
+    }
+
+    if(!wglShareLists(hGLRC_pbuf, hGLRC))
+    {
+        printf("Error: wglShareLists failed\n");
+        exit(1);
+    }
+#else // _WIN32
    int scrnum;
    GLXFBConfig *fbconfig;
    XVisualInfo *visinfo;
@@ -758,7 +1021,11 @@ static void make_pbuffer(int width, int height)
    }
 
    pbuf = glXCreatePbuffer(dpy, fbconfig[0], pbufAttrib);
-
+   if( !pbuf )
+   {
+      fprintf(stderr, "Error: couldn't create requested pbuffer\n");
+      exit(1);
+   }
    visinfo = glXGetVisualFromFBConfig(dpy, fbconfig[0]);
    if (!visinfo) {
       fprintf(stderr, "Error: couldn't get an RGBA, double-buffered visual\n");
@@ -773,12 +1040,29 @@ static void make_pbuffer(int width, int height)
 
    XFree(fbconfig);
    XFree(visinfo);
+#endif //_WIN32
 }
 
 
 static void event_loop(void)
 {
-   while (1) {
+    while (1)
+    {
+#ifdef _WIN32
+        MSG msg;
+        if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))   
+        {
+            if (msg.message==WM_QUIT)               
+            {
+                break;
+            }
+            else
+            {
+                TranslateMessage(&msg);             // Translate The Message
+                DispatchMessage(&msg);              // Dispatch The Message
+            }
+        }
+#else // _WIN32
       while (XPending(dpy) > 0) {
          XEvent event;
          XNextEvent(dpy, &event);
@@ -822,6 +1106,7 @@ static void event_loop(void)
             }
          }
       }
+#endif //_WIN32
 
       /* next frame */
       angle += 2.0;
@@ -869,6 +1154,18 @@ static void event_loop(void)
 static 
 void my_print_usage(void)
 {
+#ifdef _WIN32
+    fprintf(stdout,
+           "\n"
+           "Usage:\n"
+           "------\n"
+           "\n"
+           "-loop <n>         loop demo n frames then exit, n > 0\n"
+           "-help             print this information\n"
+           "-info             print OpenGL information (vendor etc.) to stdout\n"
+           "-fbo              use GL_EXT_framebuffer_object instead of WGL_ARB_pbuffer\n"
+           "\n");
+#else // _WIN32
     fprintf(stdout,
            "\n"
            "Usage:\n"
@@ -880,6 +1177,7 @@ void my_print_usage(void)
            "-info             print GLX information (vendor etc.) to stdout\n"
            "-fbo              use GL_EXT_framebuffer_object instead of GLX_SGIX_pbuffer\n"
            "\n");
+#endif // _WIN32
 }
 
 
@@ -887,11 +1185,18 @@ int
 main(int argc, char *argv[])
 {
    int i;
+#ifndef _WIN32
    char *dpyName       = NULL;
+#endif // _WIN32
    GLboolean printInfo = GL_FALSE;
 
    /* parse arguments */
-   if (argc > 6) {
+#ifdef _WIN32
+   if (argc > 5)
+#else // _WIN32
+   if (argc > 6)
+#endif // _WIN32
+   {
        fprintf(stdout,
                "%s: wrong number of arguments specified ...\n",argv[0]);
        my_print_usage();
@@ -899,7 +1204,9 @@ main(int argc, char *argv[])
    }
 
    for (i = 1; i < argc; i++) {
-      if (strcmp(argv[i], "-display") == 0) {
+#ifndef _WIN32
+      if (strcmp(argv[i], "-display") == 0)
+      {
          i++;
          if (i >= argc) {
             my_print_usage();
@@ -907,7 +1214,9 @@ main(int argc, char *argv[])
          }
          dpyName = argv[i];
       }
-      else if (strcmp(argv[i], "-loop") == 0) {
+      else
+#endif // _WIN32
+      if (strcmp(argv[i], "-loop") == 0) {
          i++;
          if (i >= argc) {
             my_print_usage();
@@ -935,12 +1244,15 @@ main(int argc, char *argv[])
       }
    }
 
+#ifdef _WIN32
+#else // _WIN32
    /* open the display */
    dpy = XOpenDisplay(dpyName);
    if (!dpy) {
       printf("Error: couldn't open display %s\n", dpyName);
       return -1;
    }
+#endif // _WIN32
 
    /* setup and initial window, pbuffer and texture */
    if (use_fbo)
@@ -950,11 +1262,18 @@ main(int argc, char *argv[])
    }
    else
    {
+#ifdef _WIN32
+      printf("Using WGL_ARB_pbuffer\n");
+#else // _WIN32
       printf("Using GLX_SGIX_pbuffer\n");
+#endif // _WIN32
       make_window("PBuffer GLXGears", 0, 0, 600, 600);
    }
    reshapeCube(600, 600);
+#ifdef _WIN32
+#else // _WIN32
    XMapWindow(dpy, win);
+#endif // _WIN32
 
    if (printInfo) {
       printf("GL_RENDERER   = %s\n", (char *) glGetString(GL_RENDERER));
@@ -984,15 +1303,27 @@ main(int argc, char *argv[])
    }
    else
    {
+#ifdef _WIN32
+        wglDeleteContext(hGLRC_pbuf);
+        wglReleasePbufferDCARB(pbuf, hDC_pbuf);
+        wglDestroyPbufferARB(pbuf);
+#else // _WIN32
       glXDestroyContext(dpy, ctx_pbuf);
       glXDestroyPbuffer(dpy, pbuf);
+#endif
    }
 
    /* clean up */
+#ifdef _WIN32
+    wglDeleteContext(hGLRC);
+    ReleaseDC(hWnd, hDC);
+    DestroyWindow(hWnd);
+#else // _WIN32
    glXDestroyContext(dpy, ctx_win);
    XDestroyWindow(dpy, win);
 
    XCloseDisplay(dpy);
+#endif
 
    return 0;
 }
