@@ -768,7 +768,7 @@ extern  long ATI_API_CALL KAS_ScheduleTimeout(long n_jiffies);
 **  Interface layer to asyncIO layer
 **
 *******************************************************************************/
-#define FIREGL_ASYNCIO_MAX_DEV      4    /* Currently hardcode to 4  */
+#define FIREGL_ASYNCIO_MAX_DEV      32    /* Currently hardcode to 32  */
 #define FIREGL_ASYNCIO_MAX_FILE     32   /* Currently hardcode to 32 */
 #define FIREGL_ASYNCIO_MAX_SEMA     ((FIREGL_ASYNCIO_MAX_FILE+1)*FIREGL_ASYNCIO_MAX_DEV)
 
@@ -814,6 +814,11 @@ extern void ATI_API_CALL KCL_spin_unlock(void *lock);
 extern void ATI_API_CALL KCL_get_random_bytes(void *buf, int nbytes);
 extern void* ATI_API_CALL KCL_get_pubdev(void);
 
+//The length of uuid, standardized by the Open Software Foundation, is 16.
+//In kernel, hard coded in function generate_random_uuid(). 
+#define FIREGL_UUID_LEN     16
+extern void ATI_API_CALL KCL_create_uuid(void *buf);
+
 extern void ATI_API_CALL adapter_chain_init(void);
 extern void ATI_API_CALL adapter_chain_cleanup(void);
 extern void ATI_API_CALL cf_object_init(void);
